/*
LinuxCNC component for controlling the Deckel FP4 gearbox.

Copymitte (C) 2018 Sergey 'Jin' Bostandzhyan <jin@mediatomb.cc>

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
*/

#ifndef __FP4_COMMON_H__
#define __FP4_COMMON_H__

/* structure that allows to group pins together */
#define FP4_PINS_IN_GROUP    3
typedef struct
{
    hal_bit_t *p[FP4_PINS_IN_GROUP];
} pin_group_t;

/* description of a particular gear/speed setting */
typedef struct
{
    unsigned key;
    unsigned value;
} pair_t;


/* Each shaft is described by 3 bits:
 *         2       1        0
 * msb  "front" "hinten" "back" lsb
 *
 * The table below provides the mask for all three shafts.
 * The helper macros should be used on an extracted 4 bit value.
 */
#define FP4_STAGE_IS_vorne(mask)          (mask & 1)
#define FP4_STAGE_IS_mitte(mask)         ((mask >> 1) & 1)
#define FP4_STAGE_IS_hinten(mask)        ((mask >> 2) & 1)


/* lookup table from rpm to gearbox status pin values */
static pair_t FP4_gears[] =
						/*  s/f	|Block1 |   |Block2 |   | Vorge.|    */
{  /* rpm   bitmask    msb  9   8   7   6   5   4   3   2   1   0  lsb  Bit 9 is fast or slow	*/ 
	{	0,505	 },		/*	0	1	1	1	1	1	1	0	0	1	*/
	{	40,74	 },		/*	0	0	0	1	0	0	1	0	1	0	*/
	{	50,138	 },		/*	0	0	1	0	0	0	1	0	1	0	*/
	{	63,266	 },		/*	0	1	0	0	0	0	1	0	1	0	*/
	{	80,82	 },		/*	0	0	0	1	0	1	0	0	1	0	*/
	{	100,650  },		/*	1	0	1	0	0	0	1	0	1	0	*/
	{	125,778	 },		/*	1	1	0	0	0	0	1	0	1	0	*/
	{	160,594  },		/*	1	0	0	1	0	1	0	0	1	0	*/
	{	200,658  },		/*	1	0	1	0	0	1	0	0	1	0	*/
	{	250,786	 },		/*	1	1	0	0	0	1	0	0	1	0	*/
	{	315,610  },		/*	1	0	0	1	1	0	0	0	1	0	*/
	{	400,674  },		/*	1	0	1	0	1	0	0	0	1	0	*/
	{	500,802	 },		/*	1	1	0	0	1	0	0	0	1	0	*/
	{	630,1114 },		/*	1	0	0	1	0	0	1	1	0	0	*/
	{	800,1214 },		/*	1	0	1	0	0	0	1	1	0	0	*/
	{	1000,1414},		/*	1	1	0	0	0	0	1	1	0	0	*/
	{	1250,1124},		/*	1	0	0	1	0	1	0	1	0	0	*/
	{	1600,1224},		/*	1	0	1	0	0	1	0	1	0	0	*/
	{	2000,788 },		/*	1	1	0	0	0	1	0	1	0	0	*/
	{	2500,612 },		/*	1	0	0	1	1	0	0	1	0	0	*/
	{	3150,676 },		/*	1	0	1	0	1	0	0	1	0	0	*/
	{	2500,804 },		/*	1	1	0	0	1	0	0	1	0	0	*/
};

/* Furthest CCW position, marked as "red" on the Deckel   */
#define FP4_STAGE_POS_vorne               1   /* 001 */

/* block2 shaft position, marked as "bloue" on the Deckel */
#define FP4_STAGE_POS_hinten             2   /* 010 */

/* Furthest CW position, marked as "yellow" on the Deckel */
#define FP4_STAGE_POS_mitte              4   /* 100 */

/* total number of selectable gears including neutral */
#define FP4_NUM_GEARS        (sizeof(FP4_gears)/sizeof(pair_t))
/* max gear index in array */
#define FP4_MAX_GEAR_INDEX       FP4_NUM_GEARS - 1
/* index of neutral gear */
#define FP4_NEUTRAL_GEAR_INDEX   0
/* min spindle rpm > 0 supported by the FP4 */
#define FP4_MIN_RPM              80
/* index of the first non 0 rpm setting in the gears array */
#define FP4_MIN_RPM_INDEX        1
/* max spindle rpm supported by the FP4 */
#define FP4_MAX_RPM              4000

#define FP4_TWITCH_KEEP_PIN_ON   800*1000000L /* 800ms in nanoseconds */
#define FP4_TWITCH_KEEP_PIN_OFF  200*1000000L /* 200ms in nanoseconds */

/* When shifting, poll the stage pins each 5ms. Picking a lower value here
 * to make sure that we do not "miss" and do not overshoot our target
 * position. */
#define FP4_GEAR_STAGE_POLL_INTERVAL 5*1000000L /* 5ms in nanoseconds */

/* If reverse direction needs to be activated, we have to wait 100ms before
 * we activate the motor after the reverse pin has been activated or
 * deactivated */
#define FP4_REVERSE_MOTOR_INTERVAL   100*1000000L /* 100ms in nanoseconds */

/* Interval between all remaining pin operations related to gear shifting */
#define FP4_GENERIC_PIN_INTERVAL     100*1000000L /* 100ms in nanoseconds */

/* TODO: make this a module parameter */
#define FP4_WAIT_SPINDLE_AT_SPEED    500*1000000L /* 500ms in nanoseconds */
/* generic state function */
typedef void (*statefunc)(long period);

#endif//__FP4_COMMON_H__
